/*
Copyright 2016 - 2017 Huawei Technologies Co., Ltd. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package utils

import (
	"bytes"
	"crypto"
	"crypto/md5"
	"crypto/rand"
	"crypto/rsa"
	"crypto/sha256"
	"crypto/sha512"
	"crypto/x509"
	"encoding/base64"
	"encoding/hex"
	"encoding/json"
	"encoding/pem"
	"errors"
	"fmt"
	"io"
	"net"
	"os"
	"reflect"
	"regexp"
	"strings"
	"time"

	"io/ioutil"
	"log"
	"path/filepath"

	"github.com/fernet/fernet-go"
)

type EncryptMethod string

const (
	EncryptRSA          = "rsa"
	EncryptNone         = "none"
	EncryptNotSupported = "not-supported"
)

//NewEncryptMethod
func NewEncryptMethod(method string) EncryptMethod {
	switch method {
	case string(EncryptRSA):
		return EncryptRSA
	case "":
		return EncryptNone
	case string(EncryptNone):
		return EncryptNone
	}

	return EncryptNotSupported
}

//IsDirExist checks if a path is an existed dir
func IsDirExist(path string) bool {
	fi, err := os.Stat(path)

	if err != nil {
		return os.IsExist(err)
	}

	return fi.IsDir()
}

//IsFileExist checks if a file url is an exist file
func IsFileExist(filename string) bool {
	_, err := os.Stat(filename)
	return err == nil || os.IsExist(err)
}

//Contain checks if a target item exists in the list/map
func Contain(obj interface{}, target interface{}) (bool, error) {
	targetValue := reflect.ValueOf(target)

	switch reflect.TypeOf(target).Kind() {
	case reflect.Slice, reflect.Array:
		for i := 0; i < targetValue.Len(); i++ {
			if targetValue.Index(i).Interface() == obj {
				return true, nil
			}
		}
	case reflect.Map:
		if targetValue.MapIndex(reflect.ValueOf(obj)).IsValid() {
			return true, nil
		}
	}

	return false, errors.New("not in array")
}

//ValidatePassword verifies if a password is valid
func ValidatePassword(password string) error {
	if valida, _ := regexp.MatchString("[:alpha:]", password); valida != true {
		return fmt.Errorf("no alpha character in the password")
	}

	if valida, _ := regexp.MatchString("[:digit:]", password); valida != true {
		return fmt.Errorf("no digital character in the password")
	}

	if len(password) < 5 || len(password) > 30 {
		return fmt.Errorf("password characters length should be between 5 - 30")
	}

	return nil
}

//EncodeBasicAuth encode username and password into a base64 string
func EncodeBasicAuth(username string, password string) string {
	auth := username + ":" + password
	msg := []byte(auth)
	authorization := make([]byte, base64.StdEncoding.EncodedLen(len(msg)))
	base64.StdEncoding.Encode(authorization, msg)
	return string(authorization)
}

//DecodeBasicAuth decode a base64 string into a username and a password
func DecodeBasicAuth(authorization string) (username string, password string, err error) {
	basic := strings.Split(strings.TrimSpace(authorization), " ")
	if len(basic) <= 1 {
		return "", "", err
	}

	decLen := base64.StdEncoding.DecodedLen(len(basic[1]))
	decoded := make([]byte, decLen)
	authByte := []byte(basic[1])
	n, err := base64.StdEncoding.Decode(decoded, authByte)

	if err != nil {
		return "", "", err
	}
	if n > decLen {
		return "", "", fmt.Errorf("something went wrong decoding auth config")
	}

	arr := strings.SplitN(string(decoded), ":", 2)
	if len(arr) != 2 {
		return "", "", fmt.Errorf("invalid auth configuration file")
	}

	username = arr[0]
	password = strings.Trim(arr[1], "\x00")

	return username, password, nil
}

//MD5 generates a md value of a key
func MD5(key string) string {
	md5String := fmt.Sprintf("dockyard %s is a container %d hub", key, time.Now().Unix())
	h := md5.New()
	h.Write([]byte(md5String))

	return hex.EncodeToString(h.Sum(nil))
}

//GenerateRSAKeyPair generate a private key and a public key
func GenerateRSAKeyPair(bits int) ([]byte, []byte, error) {
	privKey, err := rsa.GenerateKey(rand.Reader, bits)
	if err != nil {
		return nil, nil, err
	}

	privBytes := x509.MarshalPKCS1PrivateKey(privKey)
	pubBytes, err := x509.MarshalPKIXPublicKey(&privKey.PublicKey)

	if err != nil {
		return nil, nil, err
	}

	privBlock := &pem.Block{Type: "RSA PRIVATE KEY", Bytes: privBytes}
	pubBlock := &pem.Block{Type: "RSA PUBLIC KEY", Bytes: pubBytes}

	return pem.EncodeToMemory(privBlock), pem.EncodeToMemory(pubBlock), nil
}

//RSAEncrypt encrypts a content by a public key
func RSAEncrypt(keyBytes []byte, contentBytes []byte) ([]byte, error) {
	pubKey, err := getPubKey(keyBytes)
	if err != nil {
		return nil, err
	}

	return rsa.EncryptPKCS1v15(rand.Reader, pubKey, contentBytes)
}

//RSADecrypt decrypts content by a private key
func RSADecrypt(keyBytes []byte, contentBytes []byte) ([]byte, error) {
	privKey, err := getPrivKey(keyBytes)
	if err != nil {
		return nil, err
	}

	return rsa.DecryptPKCS1v15(rand.Reader, privKey, contentBytes)
}

//SHA256Sign signs a content by a private key
func SHA256Sign(keyBytes []byte, contentBytes []byte) ([]byte, error) {
	privKey, err := getPrivKey(keyBytes)
	if err != nil {
		return nil, err
	}

	hashed := sha256.Sum256(contentBytes)
	return rsa.SignPKCS1v15(rand.Reader, privKey, crypto.SHA256, hashed[:])
}

//SHA256Verify verifies if a content is valid by a signed data an a public key
func SHA256Verify(keyBytes []byte, contentBytes []byte, signBytes []byte) error {
	pubKey, err := getPubKey(keyBytes)
	if err != nil {
		return err
	}

	signStr := hex.EncodeToString(signBytes)
	newSignBytes, _ := hex.DecodeString(signStr)
	hashed := sha256.Sum256(contentBytes)
	return rsa.VerifyPKCS1v15(pubKey, crypto.SHA256, hashed[:], newSignBytes)
}

//getPrivKey
func getPrivKey(privBytes []byte) (*rsa.PrivateKey, error) {
	block, _ := pem.Decode(privBytes)
	if block == nil {
		return nil, errors.New("fail to decode private key")
	}

	return x509.ParsePKCS1PrivateKey(block.Bytes)
}

//getPubKey
func getPubKey(pubBytes []byte) (*rsa.PublicKey, error) {
	block, _ := pem.Decode(pubBytes)
	if block == nil {
		return nil, errors.New("Fail to decode public key")
	}

	pubInterface, err := x509.ParsePKIXPublicKey(block.Bytes)
	if err != nil {
		return nil, err
	}

	pubKey, ok := pubInterface.(*rsa.PublicKey)
	if !ok {
		return nil, errors.New("Fail get public key from public interface")
	}

	return pubKey, nil
}

//TokenMarshal encrypts data in `v`
func TokenMarshal(v interface{}, key string) ([]byte, error) {
	var buf bytes.Buffer
	err := json.NewEncoder(&buf).Encode(v)
	if err != nil {
		return nil, err
	}

	k, err := fernet.DecodeKey(key)
	if err != nil {
		return nil, err
	}
	return fernet.EncryptAndSign(buf.Bytes(), k)
}

//TokenUnmarshal decryptes a token and save the original data to `v`.
func TokenUnmarshal(token string, key string, v interface{}) error {
	k, err := fernet.DecodeKey(key)
	if err != nil {
		return err
	}

	msg := fernet.VerifyAndDecrypt([]byte(token), time.Hour, []*fernet.Key{k})
	if msg == nil {
		return errors.New("invalid or expired token")
	}

	return json.NewDecoder(bytes.NewBuffer(msg)).Decode(&v)
}

//GetFileSize get the size(bytes) of file.
func GetFileSize(path string) (int64, error) {
	if file, err := os.Open(path); err != nil {
		return 0, err
	} else {
		header := make([]byte, 512)
		file.Read(header)
		stat, _ := file.Stat()

		defer file.Close()

		return stat.Size(), nil
	}
}

//GetFileSHA512
func GetFileSHA512(path string) (string, error) {
	if file, err := os.Open(path); err != nil {
		return "", err
	} else {
		hs := sha512.New()
		io.Copy(hs, file)

		return fmt.Sprintf("%x", hs.Sum(nil)), nil
	}
}

const (
	letterBytes   = "abcdefghijklmnopqrstuvwxyz" // 52 possibilities
	letterIdxBits = 6                            // 6 bits to represent 64 possibilities / indexes
	letterIdxMask = 1<<letterIdxBits - 1         // All 1-bits, as many as letterIdxBits
)

//RandomString
func RandomString(length int) string {
	result := make([]byte, length)
	bufferSize := int(float64(length) * 1.3)

	for i, j, randomBytes := 0, 0, []byte{}; i < length; j++ {
		if j%bufferSize == 0 {
			randomBytes = SecureRandomBytes(bufferSize)
		}
		if idx := int(randomBytes[j%length] & letterIdxMask); idx < len(letterBytes) {
			result[i] = letterBytes[idx]
			i++
		}
	}

	return string(result)
}

//SecureRandomBytes returns the requested number of bytes using crypto/rand
func SecureRandomBytes(length int) []byte {
	var randomBytes = make([]byte, length)
	_, err := rand.Read(randomBytes)
	if err != nil {
		log.Fatal("Unable to generate random bytes")
	}
	return randomBytes
}

//ListDir
func ListDir(dirPth string, suffix string) (files []string, err error) {
	files = make([]string, 0, 10)

	dir, err := ioutil.ReadDir(dirPth)
	if err != nil {
		return nil, err
	}

	PthSep := string(os.PathSeparator)
	suffix = strings.ToUpper(suffix)

	for _, fi := range dir {
		if fi.IsDir() {
			continue
		}
		if strings.HasSuffix(strings.ToUpper(fi.Name()), suffix) { //匹配文件
			files = append(files, dirPth+PthSep+fi.Name())
		}
	}

	return files, nil
}

//WalkDir
func WalkDir(dirPth, suffix string) (files []string, err error) {
	files = make([]string, 0, 30)
	suffix = strings.ToUpper(suffix)

	err = filepath.Walk(dirPth, func(filename string, fi os.FileInfo, err error) error {
		if fi.IsDir() {
			return nil
		}

		if strings.HasSuffix(strings.ToUpper(fi.Name()), suffix) {
			files = append(files, filename)
		}

		return nil
	})

	return files, err
}

// WaitForHostPort wait for the $HOST:$PORT to be ready
func WaitForHostPort(host string, port int, loopDuration time.Duration, retryTimes int) error {
	serviceReady := false
	log.Printf("waiting for %s:%d to be ready\n", host, port)
	for i := 0; i < retryTimes; i++ {
		log.Printf("ping %s:%d for the %d time\n", host, port, (i + 1))
		err := ping(host, port)
		if err == nil {
			serviceReady = true
			break
		} else {
			time.Sleep(time.Second * loopDuration)
		}
	}

	if serviceReady == true {
		return nil
	} else {
		return fmt.Errorf("timeout(more than %d seconds) waiting for service %s:%d", (int(loopDuration) * retryTimes), host, port)
	}
}

func ping(host string, port int) error {
	addr := fmt.Sprintf("%s:%d", host, port)
	_, err := net.Dial("tcp", addr)
	return err
}
