/*
Copyright 2016 - 2017 Huawei Technologies Co., Ltd. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package model

import (
	"sync"
	"time"
)

//DockerV2 save Docker image data which uploading with Docker client follow Docker Distribution protocal
type DockerV2 struct {
	ID            int64      `json:"id" gorm:"column:id;primary_key"`
	Namespace     string     `json:"namespace" sql:"not null;type:varchar(255)"  gorm:"column:namespace;unique_index:dockerv2_repository"`
	Repository    string     `json:"repository" sql:"not null;type:varchar(255)"  gorm:"column:repository;unique_index:dockerv2_repository"`
	SchemaVersion string     `json:"schema_version" sql:"not null;type:varchar(255)" gorm:"column:schema_version"`
	Manifests     string     `json:"manifests" sql:"null;type:text" gorm:"column:manifests"`
	Agent         string     `json:"agent" sql:"null;type:text" gorm:"column:agent"`
	Short         string     `json:"short" sql:"null;type:text" gorm:"column:short"`
	Description   string     `json:"description" sql:"null;type:text" gorm:"column:description"`
	Size          int64      `json:"size" sql:"default:0" gorm:"column:size"`
	Locked        bool       `json:"locked" sql:"default:false" gorm:"column:locked"`
	CreatedAt     time.Time  `json:"create_at" sql:"" gorm:"column:create_at"`
	UpdatedAt     time.Time  `json:"update_at" sql:"" gorm:"column:update_at"`
	DeletedAt     *time.Time `json:"delete_at" sql:"index" gorm:"column:delete_at"`
}

//TableName is table name of DockerV2 model
func (r *DockerV2) TableName() string {
	return "docker_v2"
}

//DockerImageV2 is
type DockerImageV2 struct {
	ID              int64      `json:"id" gorm:"primary_key" gorm:"column:id"`
	ImageID         string     `json:"image_id" sql:"null;type:varchar(255)" gorm:"column:image_id"`
	BlobSum         string     `json:"blob_sum" sql:"null;type:varchar(255)" gorm:"column:blob_sum"`
	V1Compatibility string     `json:"v1_compatibility" sql:"null;type:text" gorm:"column:v1_compatibility"`
	Path            string     `json:"path" sql:"null;type:text" gorm:"column:path"`
	OSS             string     `json:"oss" sql:"null;type:text" gorm:"column:oss"`
	Size            int64      `json:"size" sql:"default:0" gorm:"column:size"`
	Locked          bool       `json:"locked" sql:"default:false" gorm:"column:locked"`
	CreatedAt       time.Time  `json:"create_at" sql:"" gorm:"column:create_at"`
	UpdatedAt       time.Time  `json:"update_at" sql:"" gorm:"column:update_at"`
	DeletedAt       *time.Time `json:"delete_at" sql:"index" gorm:"column:delete_at"`
}

//TableName is
func (i *DockerImageV2) TableName() string {
	return "docker_image_v2"
}

//DockerTagV2 is
type DockerTagV2 struct {
	ID            int64      `json:"id" gorm:"primary_key" gorm:"column:id"`
	DockerV2      int64      `json:"docker_v2" sql:"not null;default:0" gorm:"column:docker_v2"`
	Tag           string     `json:"tag" sql:"not null;type:varchar(255)" gorm:"column:tag"`
	ImageID       string     `json:"image_id" sql:"not null;type:varchar(255)" gorm:"column:image_id"`
	Manifest      string     `json:"manifest" sql:"null;type:text" gorm:"column:manifest"`
	SchemaVersion string     `json:"schema_version" sql:"not null;type:varchar(255)" gorm:"column:schema_version"`
	CreatedAt     time.Time  `json:"create_at" sql:"" gorm:"column:create_at"`
	UpdatedAt     time.Time  `json:"update_at" sql:"" gorm:"column:update_at"`
	DeletedAt     *time.Time `json:"delete_at" sql:"index" gorm:"column:delete_at"`
}

//TableName is
func (t *DockerTagV2) TableName() string {
	return "docker_tag_v2"
}

//GetTags return tas data of repository.
func (r *DockerV2) GetTags(namespace, repository string) ([]string, error) {
	r.Namespace, r.Repository = namespace, repository

	if err := DB.Debug().Where("namespace = ? AND repository = ?", namespace, repository).First(&r).Error; err != nil {
		return []string{}, err
	}

	var tags []DockerTagV2
	result := []string{}

	if err := DB.Debug().Where("docker_v2 = ?", r.ID).Find(&tags).Error; err != nil {
		return []string{}, err
	}

	for _, tag := range tags {
		result = append(result, tag.Tag)
	}

	return result, nil

}

//Get is
func (r *DockerV2) Get(namespace, repository string) error {
	if err := DB.Debug().Where("namespace = ? AND repository =? ", namespace, repository).First(&r).Error; err != nil {
		return err
	}

	return nil
}

var dockerv2Mutex sync.Mutex

//Put is
func (r *DockerV2) Put(namespace, repository string) error {
	r.Namespace, r.Repository = namespace, repository

	dockerv2Mutex.Lock()
	defer dockerv2Mutex.Unlock()

	tx := DB.Begin()
	if err := tx.Debug().Where("namespace = ? AND repository = ? ", namespace, repository).FirstOrCreate(&r).Error; err != nil {
		tx.Rollback()
		return err
	}

	tx.Commit()
	return nil
}

//PutAgent is
func (r *DockerV2) PutAgent(namespace, repository, agent, version string) error {
	r.Namespace, r.Repository = namespace, repository

	tx := DB.Begin()

	if err := tx.Debug().Where("namespace = ? AND repository = ? ", namespace, repository).FirstOrCreate(&r).Error; err != nil {
		tx.Rollback()
		return err
	}

	if err := tx.Debug().Model(&r).Updates(map[string]string{"agent": agent, "schema_version": version}).Error; err != nil {
		tx.Rollback()
		return err
	}

	tx.Commit()
	return nil
}

//Get is
func (i *DockerImageV2) Get(blobsum string) error {
	i.BlobSum = blobsum

	if err := DB.Debug().Where("blob_sum = ?", blobsum).First(&i).Error; err != nil {
		return err
	}

	return nil
}

//Put is
func (i *DockerImageV2) Put(tarsum, path string, size int64) error {
	i.BlobSum, i.Path, i.Size = tarsum, path, size

	tx := DB.Begin()

	if err := tx.Debug().Where("blob_sum = ? ", tarsum).FirstOrCreate(i).Error; err != nil {
		tx.Rollback()
		return err
	}

	tx.Commit()
	return nil
}

//Get is get DockerTagV2 data
func (t *DockerTagV2) Get(namespace, repository, tag string) (*DockerTagV2, error) {
	r := new(DockerV2)

	if err := DB.Debug().Where("namespace = ? AND repository = ?", namespace, repository).First(&r).Error; err != nil {
		return t, err
	}

	if err := DB.Debug().Where("docker_v2 = ? AND tag = ?", r.ID, tag).First(&t).Error; err != nil {
		return t, err
	}

	return t, nil
}

//Put is
func (t *DockerTagV2) Put(namespace, repository, tag, imageID, manifest, schema string) error {
	r := new(DockerV2)

	if err := DB.Debug().Where("namespace = ? AND repository = ? ", namespace, repository).First(&r).Error; err != nil {
		return err
	}

	tx := DB.Begin()
	t.DockerV2, t.Tag, t.ImageID, t.Manifest, t.SchemaVersion = r.ID, tag, imageID, manifest, schema

	if err := tx.Debug().Where("docker_v2 = ? AND tag = ?", r.ID, tag).FirstOrCreate(&t).Error; err != nil {
		tx.Rollback()
		return err
	}

	if err := tx.Debug().Model(&t).Updates(map[string]interface{}{"image_id": imageID, "manifest": manifest, "schema_version": string(schema)}).Error; err != nil {
		tx.Rollback()
		return err
	}

	tx.Commit()
	return nil
}
